@@ -198,6 +198,19 @@ class LabApp(NotebookApp):
     watch = Bool(False, config=True,
         help="Whether to serve the app in watch mode")
 
+    def init_webapp(self, *args, **kwargs):
+        super().init_webapp(*args, **kwargs)
+        settings = self.web_app.settings
+        if 'page_config_data' not in settings:
+            settings['page_config_data'] = {}
+
+        # Handle quit button with support for Notebook < 5.6
+        if hasattr(self, 'quit_button'):
+            quit_button = self.quit_button
+        else:
+            quit_button = True
+        settings['page_config_data']['quit_button'] = quit_button
+
     def init_server_extensions(self):
         """Load any extensions specified by config.
 
